import 'dart:convert';
import 'dart:io';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:dio_logger/dio_logger.dart';
import 'package:flutter_mylibrary/config/const.dart';
import 'package:flutter_mylibrary/utils/utils.dart';
import 'exception/http_error.dart';
import 'http_data.dart';


/// 请求方法.
class Method {
  static final String get = "GET";
  static final String post = "POST";
  static final String put = "PUT";
  static final String head = "HEAD";
  static final String delete = "DELETE";
  static final String patch = "PATCH";
}


class EasyDio {

  //服务器路径
  static EasyDio _instance = EasyDio._internal();

  Dio _dio;

  factory EasyDio() => _instance;

  BaseOptions _baseOptions;

  Function _onError;

  Function _onFinish;

  Function _onStart;

  ValidateResponse _validateResponse;

  String _successCode;

  String _codeKey;

  String _msgKey;

  String _dataKey;

  Dio get dio => _dio;

  List<String> publicErrorCode = [];

  ///-----------------------------------------工具类初始化设置-------------------------------------

  EasyDio._internal();

  ///初始化公共属性
  ///
  /// [baseUrl] 地址前缀
  /// [connectTimeout] 连接超时赶时间
  /// [receiveTimeout] 接收超时赶时间
  /// [interceptors] 基础拦截器
  void init({String baseUrl,
    int connectTimeout,
    int receiveTimeout,
    String pem,
    String pKCSPath,
    String pKCSPwd,
    String codeKey = 'code',
    String msgKey = 'msg',
    String dataKey = 'data',
    String successCode = '0',
    ResponseType responseType,
    bool openLog = true,
    int retry = 3,
    String method,
    Map<String, dynamic> headers,
    Map<String, dynamic> extra,
    String contentType,
    ValidateStatus validateStatus,
    bool followRedirects,
    Function onError,
    Function onFinish,
    Function onStart,
    ValidateResponse validateResponse,
    bool enableAutoCookie = false,
    ListFormat listFormat,
    List<String> publicErrorCode,
    List<Interceptor> interceptors
  }) {

    this._validateResponse = validateResponse;
    this._onStart = onStart;
    this._onError = onError;
    this._onFinish = onFinish;
    this._successCode = successCode;
    this._codeKey = codeKey;
    this._msgKey = msgKey;
    this._dataKey = dataKey;

    method = method??Method.post;

    if(isNotEmpty(publicErrorCode)){
      this.publicErrorCode.addAll(publicErrorCode);
    }

    _baseOptions = BaseOptions(
        baseUrl: baseUrl,
        receiveTimeout: receiveTimeout,
        connectTimeout: connectTimeout,
        contentType: contentType,
        responseType: responseType??ResponseType.json,
        headers: headers??{},
        extra: extra??{},
        validateStatus: validateStatus,
        followRedirects: followRedirects,
        listFormat: listFormat,
        method: method
    );


    _dio = Dio(_baseOptions);


    if(openLog && isInDebugMode){
      _dio.interceptors.add(dioLoggerInterceptor);
    }


    if (retry != 0) {
      // dio.interceptors.add(element)
      // dio.interceptors.add(RetryInterceptor(
      //     dio: dio,
      //     options: RetryOptions(
      //       retries: retry, // Number of retries before a failure
      //       retryInterval: const Duration(seconds: 1), // Interval between each retry
      //     )
      // ));
    }



    ///-----------------------------------------证书相关信息-------------------------------------
    (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
      client.badCertificateCallback = (X509Certificate cert, String host, int port) {
        if (isNotEmpty(pem)) {
          if (cert.pem == pem) {
            // 证书一致，则放行
            return true;
          }else{
            return false;
          }
        }else{
          return true;
        }
      };
    };

    if (isNotEmpty(pKCSPath) && isNotEmpty(pKCSPwd)) {
      (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
        SecurityContext sc = new SecurityContext();
        //file为证书路径
        sc.setTrustedCertificates(pKCSPath, password: pKCSPwd);
        HttpClient httpClient = new HttpClient(context: sc);
        return httpClient;
      };
    }


  }



  ///-----------------------------------------请求方法封装-------------------------------------
  Future<HttpData<T>> request<T>({
    String url,
    String method,
    Options options,
    String cacheKey,
    bool isFormData,
    bool publicMethod,
    var data,
    String contentType,
    CancelToken cancelToken,
    Function onStart,
    Function(HttpData<T> httpData) onSuccess,
    Function(HttpData httpData) onError,
    Function onFinish,
    ValidateResponse validateResponse,
    String successCode,
    String codeKey,
    String msgKey,
    String dataKey,
    bool checkDataEmpty = true,
    ProgressCallback onReceiveProgress,
    ProgressCallback onSendProgress,
    dynamic Function(Map map) handleRes,
    HttpError Function(dynamic data) validateError,
  }) async{

    return await _handleHttpData<T, dynamic>(
        url: url,
        method: method,
        options: options,
        cacheKey: cacheKey,
        isFormData: isFormData,
        publicMethod: publicMethod,
        data: data,
        cancelToken: cancelToken,
        onStart: onStart,
        onSuccess: onSuccess,
        onError: onError,
        onFinish: onFinish,
        validateResponse: validateResponse,
        validateError: validateError,
        successCode: successCode,
        codeKey: codeKey,
        msgKey: msgKey,
        dataKey: dataKey,
        onReceiveProgress: onReceiveProgress,
        onSendProgress: onSendProgress,
        handleRes: handleRes,
        checkDataEmpty: checkDataEmpty
    );
  }


  Future<HttpData<T>> requestList<T, E>({
    String url,
    String method,
    Options options,
    String cacheKey,
    bool isFormData,
    bool publicMethod,
    var data,
    String contentType,
    CancelToken cancelToken,
    Function onStart,
    Function(HttpData<T> httpData) onSuccess,
    Function(HttpData httpData) onError,
    Function onFinish,
    ValidateResponse validateResponse,
    String successCode,
    String codeKey,
    String msgKey,
    String dataKey,
    bool checkDataEmpty = true,
    ProgressCallback onReceiveProgress,
    ProgressCallback onSendProgress,
    dynamic Function(Map map) handleRes,
    HttpError Function(dynamic data) validateError,
  }) async{


    return await _handleHttpData<T, E>(
        url: url,
        method: method,
        options: options,
        cacheKey: cacheKey,
        isFormData: isFormData,
        publicMethod: publicMethod,
        data: data,
        cancelToken: cancelToken,
        onStart: onStart,
        onSuccess: onSuccess,
        onError: onError,
        onFinish: onFinish,
        validateResponse: validateResponse,
        validateError: validateError,
        successCode: successCode,
        codeKey: codeKey,
        msgKey: msgKey,
        dataKey: dataKey,
        onReceiveProgress: onReceiveProgress,
        onSendProgress: onSendProgress,
        handleRes: handleRes,
        checkDataEmpty: checkDataEmpty
    );
  }


  ///进行httpData组装
  Future<HttpData<T>> _handleHttpData<T, E>({
    String url,
    String method,
    Options options,
    String cacheKey,
    bool isFormData,
    bool publicMethod,
    var data,
    CancelToken cancelToken,
    Function onStart,
    Function(HttpData<T> httpData) onSuccess,
    Function(HttpData httpData) onError,
    Function onFinish,
    ValidateResponse validateResponse,
    HttpError Function(dynamic data) validateError,
    String successCode,
    String codeKey,
    String msgKey,
    bool checkDataEmpty = true,
    String dataKey,
    ProgressCallback onReceiveProgress,
    ProgressCallback onSendProgress,
    dynamic Function(Map map) handleRes,
  }) async{

    HttpData httpData = HttpData<T>();

    ///httpData组装数据
    String code;
    String msg;
    dynamic resultData;


    try{
      isFormData = isFormData??false;
      publicMethod = publicMethod??true;

      if(isNotEmpty(onStart)){
        onStart.call();
      }else if(publicMethod) {
        this._onStart?.call();
      }


      Response response;

      Options mOptions = Options();

      if(isNotEmpty(options)){
        mOptions = options;
      }

      if (isFormData) {
        data = FormData.fromMap(data);
      }

      if(isNotEmpty(method)){
        mOptions.method =  method;
      }else{
        mOptions.method = _baseOptions.method;
      }


      if (method == Method.get) {
        response = await _dio.request(
            url,
            queryParameters: data,
            options: mOptions,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }else{
        response = await _dio.request(
            url,
            data: data,
            options: mOptions,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }



      String mCodeKey = codeKey??_codeKey;
      String mMsgKey = msgKey??_msgKey;
      String mDataKey = dataKey??_dataKey;
      String mSuccessCode = successCode??_successCode;


      if (response.data is List) {
        resultData = response.data;
        code = successCode??_successCode;
      }else{
        Map<String, dynamic> resJson;

        try{
          if(response.data is Map){
            resJson = response.data;
          }else{
            resJson = json.decode(response.data);
          }
        }catch(e){
          throw HttpError.parseDataError(e, logMsg: 'response.data转换json报错');
        }

        if(resJson.containsKey(mCodeKey)){
          code = resJson[mCodeKey].toString();
        }

        if(resJson.containsKey(mMsgKey)){
          msg = resJson[mMsgKey].toString();
        }


        if (isNotEmpty(validateResponse)) {
          resJson = validateResponse(response.data);
        }else if(isNotEmpty(_validateResponse)){
          resJson = _validateResponse(response.data);
        }


        if(code != mSuccessCode){
          throw HttpError(HttpErrorType.ResultError, isNotEmpty(msg)? msg : HTTP_INCORRECT_MSG, logMsg: '请求code错误，成功code: $mSuccessCode, 接口返回code: $code');
        }else{

          if(resJson.containsKey(mDataKey)){
            ///解析data
            var rawData = resJson[mDataKey];
            if(isNotEmpty(handleRes)){
              try{
                if(rawData is List){
                  resultData = List<E>.from(rawData.map((e) => handleRes(e)).toList());
                }else if(rawData is Map){
                  resultData = handleRes(rawData);
                }
              }catch(e){
                throw HttpError.parseDataError(e);
              }
            }

            if(isEmpty(resultData) && checkDataEmpty){
              throw HttpError.emptyDataError(response.data);
            }
          }else{
            if(checkDataEmpty){
              throw HttpError.emptyDataError(response.data);
            }
          }
        }
      }
    }catch(e){
      HttpError error;
      if (e is DioError) {
        error = HttpError.dioError(e);
      }else if(e is HttpError){
        error = e;
      }else{
        error = HttpError.unknown(e);
      }
      httpData.httpError = error;
    }


    ///开始组装httpData
    httpData.code = code;
    httpData.msg = msg;
    httpData.data = resultData;


    try{
      if(isNotEmpty(httpData.httpError)){
        if(isNotEmpty(validateError)){
          httpData.httpError = validateError(httpData.data);
        }

        if(onError != null){
          onError.call(httpData);
          if(publicErrorCode.contains(code)){
            if (this._onError != null) {
              this._onError.call(httpData);
            }
          }
        }else if(publicMethod) {
          this._onError?.call(httpData);
        }
      }else{
        onSuccess?.call(httpData);
      }


      if(onFinish != null){
        onFinish.call();
      }else if(publicMethod) {
        this._onFinish?.call();
      }

    }catch(e){
      httpData.httpError = HttpError.unknown(e);
    }

    return httpData;
  }





  ///-----------------------------------------上传文件方法封装-------------------------------------------------------------
  Future<HttpData> fileUpload({
    MultipartFile multipartFile,
    String url,
    String filePath,
    String key = 'file',
    String fileName,
    FormData formData,
    ProgressCallback onSendProgress,
    ProgressCallback onReceiveProgress,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = false}) async{

    MultipartFile file = multipartFile??await MultipartFile.fromFile(filePath, filename: fileName);

    return await request(
      url: url,
      method: Method.post,
      publicMethod: publicMethod,
      data: formData??FormData.fromMap({
        key: file
      }),
      onSendProgress: (int sent, int total){
        onSendProgress?.call(sent, total);
      },
      onReceiveProgress: (int sent, int total){
        onReceiveProgress?.call(sent, total);
      },
    );

  }


  Future<HttpData> filesUpload({
    List<MultipartFile> multipartFile,
    String url,
    String filePath,
    String key = 'file',
    String fileName,
    FormData formData,
    ProgressCallback onSendProgress,
    ProgressCallback onReceiveProgress,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = false}) async{


    return await request(
      url: url,
      method: Method.post,
      publicMethod: publicMethod,
      data: formData??FormData.fromMap({
        key: multipartFile
      }),
      onSendProgress: (int sent, int total){
        onSendProgress?.call(sent, total);
      },
      onReceiveProgress: (int sent, int total){
        onReceiveProgress?.call(sent, total);
      },
    );
  }


  ///-----------------------------------------下载文件方法封装-------------------------------------------------------------
  void download({
    String urlPath = '',
    String savePath = '',
    bool deleteOnError = true,
    String lengthHeader = Headers.contentLengthHeader,
    ProgressCallback onReceiveProgress,
    data,
    Map<String, dynamic> queryParameters,
    CancelToken cancelToken,
    Options options,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = true}) async{

    try{
      await _dio.download(
        urlPath,
        savePath,
        deleteOnError: deleteOnError,
        lengthHeader: lengthHeader,
        data: data,
        options: options,
        queryParameters: queryParameters,
        cancelToken: cancelToken,
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
      );
    }on DioError catch(e){
      throw HttpError.dioError(e);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (publicMethod) {
          this._onError?.call(e);
        }

        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(publicMethod){
          this._onFinish?.call();
        }
      }
    }
  }




  ///-----------------------------------------dio部分设置-------------------------------------


  void setCookie(String cookie){
    Map<String, dynamic> _headers = new Map();
    _headers["Cookie"] = cookie;
    _dio.options.headers.addAll(_headers);

  }


  void setHeaders(Map<String, dynamic> headers){
    _dio.options.headers.addAll(headers);
  }

  void removeHeaders(String key){
    _dio.options.headers.remove(key);
  }

  ///修改默认url
  void changeBaseUrl(String baseUrl){
    _dio.options.baseUrl = baseUrl;
  }

}


